/*
	Standard library file for PROGS SMS C library

	MEMHeap handling library

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 23 October 1995
	23.09.1996 MEMHeapZap(NULL) now possible
	01.08.1997 problems with MEMHeapRelease fixed
*/

#include "mem.h"
#include "err.h"
#include "job.h"

/* first some internal definitions of datatypes */

typedef struct list_ {
	struct list_ *next; 		/* next free segment */
	} list;

typedef struct _freelist {
	struct _freelist *next; 	/* next free segment */
	Size size;					/* size of this free segment */
	} freelist;

typedef struct {
	Size BlockSize; 			/* minimum size of a block */
	freelist *free; 			/* list of free areas - sorted !! */
	list *blocks;				/* list of blocks in heap */
	Job owner;					/* owner of user memory */
	} MEMHeap;

/* Heap should be defined as :
typedef MEMHeap *Heap;
	but I did not do this, because I want to hide the representation */

/* --- start of the actual routines --- */

Error MEMHeapInit(Size BlockSize, Heap *id)
{
	MEMHeap *mem;

	if (BlockSize<=0) return ERR_ORNG;
	if (!MEMAllocate(sizeof(MEMHeap), &mem))
	{
		mem->BlockSize=BlockSize;
		JOBMe(&mem->owner);

		*id=(Heap)mem;
		return ERR_OK;
	}
	else return ERR_IMEM;
}

Error MEMHeapAllocate(Heap id, Size size, void **base)
{
	MEMHeap *mem=(MEMHeap *)id;
	freelist *free, *last=NULL;
	Error err=ERR_OK;

	size=(size+sizeof(Size)+sizeof(freelist)-1)
			/sizeof(freelist)*sizeof(freelist);

	for (free=mem->free ; free ; last=free, free=free->next)
		if (free->size>=size) break;

	if (free)
	{
		/* use area in this free element */
		Size left=free->size-size;

		if (left>=sizeof(freelist))
		{
			/* link new free element */
			freelist *new;

			new=(freelist *)(((char *)free)+size);
			new->size=left;
			new->next=free->next;
			if (!last)
				mem->free=new;
			else
				last->next=new;
		}
		else
		{
			/* unlink element */
			if (!last)
				mem->free=free->next;
			else
				last->next=free->next;
		}

		/* prepare base for return */
		*((Size *)free)=size;
		*base=((char *)free)+sizeof(Size);
	}
	else
	{
		/* allocate new block and use that */
		Size len;
		list *block;
		Size *rel;
		
		/* is a normal block large enough ?? */
		if (size<=mem->BlockSize)
			len=mem->BlockSize;
		else
			len=size;

		/* allocate and link */
		err=MEMAllocateFor(len+sizeof(list), mem->owner, &block);
		if (err) return err;
		block->next=mem->blocks;
		mem->blocks=block;

		/* link in the free space */
		rel=(Size *)(((char *)block)+sizeof(list));
		*rel++=len;
		MEMHeapRelease(id,rel);
		return MEMHeapAllocate(id,size-sizeof(Size),base);
	}
	return err;
}

Error MEMHeapRelease(Heap id, void *base)
{
	MEMHeap *mem=(MEMHeap *)id;
	char *start, *end;
	freelist *free, *last, *curr;
	Size size;

	if (!base) return ERR_OK;

	start=(char *)base-sizeof(Size);
	size=*((Size *)start);
	end=size+start;

	/* search for correct place in free list */
	if (mem->free==NULL || (char *)mem->free > start)
	{
		free=(freelist *)start;
		/* link in at start of free list */
		/* could still need join with following block */
		if (end == (char *)mem->free)
		{
			/* join blocks */
			free->next=mem->free->next;
			free->size=size+mem->free->size;
			mem->free=free;
		}
		else
		{
			free->next=mem->free;
			free->size=size;
			mem->free=free;
		}
	}
	else
	{
		for (free=mem->free ; free && (char *)free<start ;
			 last=free, free=free->next);
		if (last->size+(char *)last == start)
		{
			/* join with previous block */
			/* see if the following block matches as well */
			if (end == (char *)free)
			{
				last->next=free->next;
				last->size+=size+free->size;
			}
			else
			{
				last->size+=size;
			}
		}
		else
		{
			curr=(freelist *)start;
			if (end == (char *)free)
			{
				curr->size=size+free->size;
				curr->next=free->next;
			}
			else
			{
				curr->size=size;
				curr->next=free;
			}
			last->next=curr;
		}
	}
	return ERR_OK;
}

Error MEMHeapZap(Heap id)
{
	MEMHeap *mem=(MEMHeap *)id;
	list *block;

	if (mem==NULL) return ERR_OK;

	/* first we have to release all the blocks */
	while (block=mem->blocks)
	{
		mem->blocks=block->next;
		MEMRelease(block);
	}

	/* now release the MEMHeap area (also handles ZapRoutine and ZapAddress) */
	return MEMRelease(mem);
}

Error MEMHeapWhenZapped(Heap id, void (*ZapRoutine)(void *base), char *ZapAddress)
{
	return MEMWhenZapped((void *)id, ZapRoutine, ZapAddress);
}
