/*
	Standard library file for PROGS SMS C library

	io handling library
	IOOpenPathFor
	IOName
	IOLength
	IOFileTypeGet
	IOFileTypeSet
	IODataSpaceGet
	IODataSpaceSet
	IOMedium
	IOMakeDirectory
	IODelete added

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 17 November 1994
	17.12.1995 added "$D" and "$P" as possibilities in path to indicate the
			   data and program directories respectively
	10.01.1996 added support to use "Global Variables" in IOOpenPath by
			   preceding var name by '$' (and inclosing in double quotes if
			   the varname contains underscores).
	11.01.1996 path can now be NULL in IOOpenPath
	15.04.1996 IOMedium adjusted to support more information (also for future)
	17.04.1996 added IOFileInfoGet and IOFileInfoPut
	23.04.1996 fixed problem with IOName when applied to 'ram1_', this fix
			   makes the call a bit more efficient when calling a directory
	16.04.1996 IOMakeDirectory no longer now returns ERR_OK when the directory
			   already existed (as a directory)
	05.05.1996 fixed IOFileInfoGet bug, cfr reading time stamps and version
	08.07.1996 IOFileInfoGet now also OK when used on a directory
	16.07.1996 recursive use of global variables now allowed in IOOpen and
			   IOOpenPath (and -For variants)
	19.07.1996 IOOpen now does not remove trailing underscore on "Global
			   Variable" replacement
	05.08.1996 IOMakeDirectory was bugged because some drivers return ERR_FDIU
			   instead of ERR_FEX when a directory already exists.
	19.09.1996 now catch a bug in the iof.xinf call on Falkenberg harddisk
			   interface, it crashed programs because too many bytes are reset
	29.09.1996 IODelete added to allow Global Variable replacement
	12.10.1996 work around a bug in some win drivers where a directory which
			   end in underscores is not created and does not give an error
	22.02.1997 IOOpenPath(For) stop scanning path when ERR_FEX occurs
	16.10.1997 IORename bug fix in code which decides whether the directory
			   has to be prepended to the given name
	17.11.1997 IOMedium did not fill in formattype properly
*/

#include "io_h"
#include "str_h"
#include "mem_h"
#include "dev_h"
#include "thing_h"

#define catch(X) if (err=(X)) return err; else

/* internal routines, data structures */

#define OPEN_DIR		(OpenKey)4
#define OPEN_FORNEWDIR	(OpenKey)97

typedef struct {
	Size length;					/* file length */
	char access;					/* file access, currently always 0 */
	FileType type;					/* file type */
	int dataspace;					/* dataspace area */
	int unused; 					/* don't ask me */
	short namelen;					/* length of filename */
	char name[36];					/* filename, not '\0' terminated */
	int noaccess_updatedate;		/* do not use this, has special calls */
	unsigned short noaccess_version;/* do not use this, has special calls */
	short dontknow; 				/* do not use this */
	int noaccess_backupdate;		/* do not use this, has special calls */
	} IOHeader;

typedef struct {
	short mediumnamelen;			/* length of mediumname string */
	char mediumname[20];			/* up to 16 chars medium name */
	short devicenamelen;			/* length of devicename string */
	char devicename[4]; 			/* up to  4 chars device name, e.g. WIN */
	char devicenumber;				/* medium number */
	char readonly;					/* TRUE if read only */
	short unitsize; 				/* allocation unit size */
	int mediumsize; 				/* total size in allocation units */
	int mediumfree; 				/* free space in allocation units */
	int headlen;					/* header length */
	char formattype;
	char formatsubtype;
	char density;
	char mediumtype;
	char removable;
	char unused[0x0f];				/* bytes set to -1 */
	char bugcatch[36];				/* catch Falkenberg harddisk driver bug */
									/* this resets 36 bytes too much - maybe */
									/* settig $32 instead of 32 .w to -1 */
	} IOInfo;

/* internal routines, prototypes */

Error _IOHeaderGet		(Channel chid, IOHeader *base);
Error _IOHeaderSet		(Channel chid, IOHeader *base);
Error _IOMedium 		(Channel chid, IOInfo *base);
Error _IORename 		(Channel chid, char *to);
Error _IOMakeDirectory	(Channel chid);
Error _IOOpenFor		(char *name, OpenKey key, Job id, Channel *ret);
Error _IODelete 		(char *name);

/* the actual routines */

#define GLOBAL_NAME "Global Variables"
#define GLOBAL_GET	0x47455420	/* "GET " */
#define GLOBAL_SET	0x53455420	/* "SET " */

Error IOOpen(char *name, OpenKey key, Channel *ret)
{
	return IOOpenFor(name,key,MYSELF,ret);
}

Error IOOpenFor(char *name, OpenKey key, Job id, Channel *ret)
{
	char string[IO_MAXFULLNAME], string2[IO_MAXFULLNAME];
	char *from, *to;
	char dir[IO_MAXFULLNAME];
	char *curr;

	curr=name;
	while(*curr=='$')
	{
		to=string;
		/* name starts with "Global Variable", so replace */
		if (curr[1]=='\"')
		{
			for (from=curr+2 ; *from && *from!='\"' ; *to++=*from++);
			if (*from) from++;
		}
		else
			for (from=curr+1 ; *from && *from!='_' ; *to++=*from++);
		*to='\0';
		if (STRSameCD(string,"D"))
			DEVDataGet(string);
		else
		if (STRSameCD(string,"P"))
			DEVProgramGet(string);
		else
		{
			if ( THINGCall(GLOBAL_NAME,GLOBAL_GET,3,string,dir,IO_MAXFULLNAME) )
				string[0]='\0';
			else
				STRCopy(dir,string);
		}
		curr=string+strlen(string)-1;
		if (*curr=='_' && *from) *curr='\0';
		STRAppendSafe(from,string,IO_MAXFULLNAME);
		STRCopy(string,string2);
		curr=string2;
	}
	if (key==OPEN_FORNEWDIR)
	{
		while (curr[strlen(curr)-1]=='_') curr[strlen(curr)-1]='\0';
		key=OPEN_NEW;
	}
	return _IOOpenFor(curr,key,id,ret);
}

Error IOOpenPathFor(char *name, OpenKey key, Job id, char *path, Channel *ret)
{
	char string[IO_MAXFULLNAME];	/* place to build full filename */
	char *curr; 					/* current position in path */
	Error err;						/* possible error */
	Size len=strlen(name)+1;		/* length of name, small optimisation */

	switch ((int)path)				/* fill in path when only a key is given */
	{
	case (int)PATH_DATA:			/* search on data default */
		path="$D";
		break;
	case (int)PATH_PROG:			/* search on program default */
		path="$P";
		break;
	case (int)PATH_DATAPROG:		/* search data first, then program default */
		path="$D;$P";
		break;
	case (int)PATH_PROGDATA:		/* search program first, then data default */
		path="$P;$D";
		break;
	}

	/* first try to open the file "as is" */
	err=IOOpenFor(name, key, id, ret);
	if (!path) return err;
	while (err && err!=ERR_FEX && *path!='\0')
	{
		/* start forming a full filename */
		char *curr=string;
		while(*path!=';' && *path!='\0')
			*curr++=*path++;
		if (*path==';') path++;
		if (*(curr-1)!='_') *curr++='_';
		*curr++='\0';
		STRAppendSafe(name,string,IO_MAXFULLNAME);
		err=IOOpenFor(string, key, id, ret);
	}
	return err;
}

Error IOName(Channel id, char *name, char *directory)
{
	char fullname[IO_MAXFULLNAME];	/* string to build name in */
	char _dummy[IO_MAXFULLNAME];	/* copy to allow NULL parameters */
	char *end;						/* end of possible directory */
	char *start;					/* start of directory without device */
	MediumInfo info;				/* for device name */
	IOHeader head;					/* for filename */
	Error err;						/* possible error */
	Channel dirchid;				/* to find directory */
	int fnlen;						/* length of fullname -1 */
	char keepc; 					/* keep for open dir */

	if (!name) name=_dummy;
	if (!directory) directory=_dummy;

	err=IOMedium(id, &info);		/* need the device name */
	if (err) return err;			/* oops probably not a directory driver */
	err=_IOHeaderGet(id, &head);	/* need file name (excluding device) */
	if (err) return err;			/* oops probably not a directory driver */
	STRCopy(info.devicename,fullname);
	STRCopy(info.devicename,directory);
	head.name[head.namelen]='\0';	/* need an eos marker in filename */
	STRAppend(head.name,fullname);	/* build full filename */

	if (head.type==FILETYPE_DIRECTORY)
	{
		/* shortcut if this is a directory */
		STRCopy(fullname,directory);
		*name='\0';
		return ERR_OK;
	}

	/* we now have the full filename, so build the directory name */
	fnlen=strlen(fullname)-1;		/* get length of fullname */
	keepc=fullname[fnlen];			/* keep last char in store */
	fullname[fnlen]='\0';			/* make one char shorter to avoid ERR_FDIU */
	catch( IOOpen(fullname, OPEN_DIR, &dirchid) );
	fullname[fnlen]=keepc;			/* restore fullname to correct one */
	catch( _IOHeaderGet(dirchid, &head) );	/* get directory name */
	err=IOClose(dirchid);			/* close directory */
									/* append '_' and set eos */
	if (head.namelen)				/* only when something to add */
	{
		/* complete directory name */
		MEMCopy(2,"_",head.name+head.namelen);
		STRAppend(head.name,directory);
	}
									/* and set filename */
	STRCopy(fullname+strlen(directory),name);

	return err;
}

Error IOLength(Channel id, Size *length)
{
	IOHeader head;
	Error err;

	err=_IOHeaderGet(id, &head);
	*length=head.length;
	return err;
}

Error IOFileTypeGet(Channel id, FileType *type)
{
	IOHeader head;
	Error err;

	err=_IOHeaderGet(id, &head);
	*type=head.type;
	return err;
}

Error IOFileTypeSet(Channel id, FileType type)
{
	IOHeader head;
	Error err;

	catch( _IOHeaderGet(id, &head) );
	head.type=type;
	err=_IOHeaderSet(id, &head);
	return err;
}

Error IODataSpaceGet(Channel id, Size *result)
{
	IOHeader head;
	Error err;

	err=_IOHeaderGet(id, &head);
	*result=head.dataspace;
	return err;
}

Error IODataSpaceSet(Channel id, Size dataspace)
{
	IOHeader head;
	Error err;

	catch( _IOHeaderGet(id, &head) );
	head.dataspace=dataspace;
	err=_IOHeaderSet(id, &head);
	return err;
}

Error IOFileInfoGet(Channel file, FileInfo *info)
{
	IOHeader header;
	Error err;

	catch( _IOHeaderGet(file, &header) );
	info->length=	 header.length;
	info->type= 	 header.type;
	info->dataspace= header.dataspace;
	if (header.type!=FILETYPE_DIRECTORY)
	{
		catch( IOBackupTimeGet(file,&info->BackupTime) );
		catch( IOUpdateTimeGet(file,&info->UpdateTime) );
		catch( IOVersionGet(file,&info->version) );
	}
	return IOName(file,info->name,NULL);
}

Error IOFileInfoSet(Channel file, FileInfo *info)
{
	IOHeader header;
	Error err;

	catch( _IOHeaderGet(file, &header) );
	header.length			  =info->length;
	header.type 			  =info->type;
	header.dataspace		  =info->dataspace;
	catch( _IOHeaderSet(file, &header) );
	catch( IOBackupTimeSet(file,info->BackupTime) );
	catch( IOUpdateTimeSet(file,info->UpdateTime) );
	err=IOVersionSet(file,info->version);
	return err;
}

Error IOMedium(Channel id, MediumInfo *info)
{
	Error err;
	IOInfo inf;

	catch( _IOMedium(id, &inf) );
	info->mediumname[inf.mediumnamelen]='\0';
	MEMCopy(inf.mediumnamelen,inf.mediumname,info->mediumname);
	info->devicename[inf.devicenamelen+1]='\0';
	info->devicename[inf.devicenamelen]=inf.devicenumber+'0';
	MEMCopy(inf.devicenamelen,inf.devicename,info->devicename);
	MEMCopy(2,"_",info->devicename+inf.devicenamelen+1);
	info->readonly	=inf.readonly;
	info->unitsize	=inf.unitsize;
	info->mediumsize=inf.mediumsize*inf.unitsize;
	info->mediumfree=inf.mediumfree*inf.unitsize;
	info->formattype=inf.formattype;
	info->formatsubtype=inf.formatsubtype;
	info->density	=inf.density;
	info->mediumtype=inf.mediumtype;
	info->removable =inf.removable;
	MEMCopy(14,&inf.unused,info->unused);
	return err;
}

Error IOMakeDirectory(char *directory)
{
	Error err;
	Channel chid;

	err=IOOpen(directory, OPEN_FORNEWDIR, &chid);
	if (err==ERR_FEX || err==ERR_FDIU || err==ERR_IPAR)
	{
		FileType type=FILETYPE_NORMAL;

		catch( IOOpen(directory,OPEN_READ,&chid) );
		IOFileTypeGet(chid,&type);
		IOClose(chid);
		if (type!=FILETYPE_DIRECTORY) return ERR_FEX;
								 else return ERR_OK;
	}
	if (err) return err;
	catch( _IOMakeDirectory(chid) );
	return IOClose(chid);
}

Error IORename(Channel id, char *to)
{
	char dir[IO_MAXDIRECTORY];
	char copy[IO_MAXFULLNAME];
	Error err;

	catch( IOName(id, NULL, dir) ); /* get filename */
	STRCopySafe(to, copy, IO_MAXFULLNAME);	/* make a copy */
	copy[strlen(dir)]='\0'; 		/* prepare for compare */
	if (!STRSameCI(copy, dir))
	{
		STRCopy(dir, copy); 		/* start with proper directory */
		STRAppend(to, copy);		/* append new name */
		return _IORename(id, copy); /* change name */
	}
	else
		return _IORename(id, to);	/* new name ok, just change it */
}

Error IODelete(char *name)
{
	char fname[IO_MAXFULLNAME], fnam2[IO_MAXFULLNAME];
	char *from, *to, *curr;

	STRCopySafe(name,fname,IO_MAXFULLNAME);
	while (fname[0]=='$')
	{
		STRCopy(fname,fnam2);
		to=fname;
		/* name starts with "Global Variable", so replace */
		if (fname[1]=='\"')
		{
			for (from=fnam2+2 ; *from && *from!='\"' ; *to++=*from++);
			if (*from) from++;
		}
		else
			for (from=fnam2+1 ; *from && *from!='_' ; *to++=*from++);
		*to='\0';
		if (STRSameCD(fname,"D"))
			DEVDataGet(fname);
		else
		if (STRSameCD(fname,"P"))
			DEVProgramGet(fname);
		else
		{
			if ( THINGCall(GLOBAL_NAME,GLOBAL_GET,3,fname,fname,IO_MAXFULLNAME) )
				fname[0]='\0';
		}
		curr=fname+strlen(fname)-1;
		if (*curr=='_' && *from) *curr='\0';
		STRAppendSafe(from,fname,IO_MAXFULLNAME);
	}
	return _IODelete(fname);
}
