/*
	Standard library file for PROGS SMS C library

	thing handling library

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera, started 4 December 1994
*/

#include "thing_h"
#include "rout_h"
#include "mem_h"
#include "str_h"

#define THING_FLAG 0x54484725		/* "THG%", the thing header flag */
#define THING_EXTE 0x00000003		/* type for an extension */
#define THING_LIST 0x01000000		/* flag to indicate list of things */
#define THING_CODE 0x18 			/* start of extension code */
#define JMP_L 0x4ef9				/* opcode for jmp.l */
#define MOVE_L_A1_SP_JSR_L 0x2f094ec9
#define ADDQ_L_4_SP_RTS    0x588f4e75

typedef struct {
	int flag, type, next, extension, def, desc;
	char code[6];
	} Extension;

typedef struct {
	int flag, type, next, extension, def, desc;
	char code[12];
	} ExtensionBlock;

typedef struct {
	int next, usage, frfre, frzap;
	Thing thing;
	int use, free, ffree, remov;
	char nshar, spare;
	int version;
	short namelen;
	char name[THING_MAXNAME];
	Error (*use_code)();
	void (*free_code)();
	void (*zap_code)();
	void (*remove_code)();
	} ThingHeader;

Error THINGLink(ThingHeader *thing);
Error SYSAllocate(Size size, Job job, void **base);
Error SYSRelease(void *base);

Error Create(ThingDefinition *def, Extension **base)
{
	int i, count;
	Extension *thingbase, *thingbase2;
	Error err;

	/* start by counting the number of extensions */
	for(count=0;def->list[count].extension;count++);

	/* now allocate and fill in the actual thing */
	err=MEMAllocateFor(sizeof(Extension)*count,def->owner,&thingbase2);
	thingbase=thingbase2;	/* to allow use of register - stupid compiler */
	if (err) return err;
	*base=thingbase;
	for (i=0;i<count;i++,thingbase++)
	{
		thingbase->flag=THING_FLAG;
		thingbase->type=THING_EXTE|THING_LIST;
		thingbase->next=sizeof(Extension);
		thingbase->extension=def->list[i].extension;
		thingbase->def=thingbase->desc=0;
		*(short *)thingbase->code=JMP_L;
		*(int *)&(thingbase->code[2])=(int)(def->list[i].function);
	}
	thingbase--;
	thingbase->type=THING_EXTE; 	/* oops, this was the last one */
	thingbase->next=0;
	
	return ERR_OK;
}

Error CreateBlock(ThingDefinition *def, Extension **base)
{
	int i, count;
	ExtensionBlock *thingbase, *thingbase2;
	Error err;

	/* start by counting the number of ExtensionBlocks */
	for(count=0;def->list[count].extension;count++);

	/* now allocate and fill in the actual thing */
	err=MEMAllocateFor(sizeof(ExtensionBlock)*count,def->owner,&thingbase2);
	thingbase=thingbase2;	/* to allow use of register - stupid compiler */
	if (err) return err;
	*base=(Extension *)thingbase;
	for (i=0;i<count;i++,thingbase++)
	{
		thingbase->flag=THING_FLAG;
		thingbase->type=THING_EXTE|THING_LIST;
		thingbase->next=sizeof(ExtensionBlock);
		thingbase->extension=def->list[i].extension;
		thingbase->def=thingbase->desc=0;
		*(int *)&(thingbase->code[0])=MOVE_L_A1_SP_JSR_L;
		*(int *)&(thingbase->code[4])=(int)def->list[i].function;
		*(int *)&(thingbase->code[8])=ADDQ_L_4_SP_RTS;
	}
	thingbase--;
	thingbase->type=THING_EXTE;   /* oops, this was the last one */
	thingbase->next=0;

	return ERR_OK;
}

static Error DoCreate(ThingDefinition *def, Error (*create)(ThingDefinition *def, Extension **base))
{
	Error err;
	Extension *thingbase;
	ThingHeader *thing, *thing2;

	err=create(def,&thingbase); 	/* create the thing itself */
	if (err) return err;

	/* now allocate and fill in the ThingHeader */
	err=SYSAllocate(sizeof(ThingHeader),def->owner,&thing2);
	thing=thing2;	/* to allow use of register - stupid compiler */
	if (err)
	{
		MEMRelease(thingbase);
		return err;
	}
	thing->namelen=min(THING_MAXNAME,strlen(def->name));
	MEMCopy(THING_MAXNAME,def->name,thing->name);
	thing->use_code=def->use_code;
	thing->free_code=def->free_code;
	thing->zap_code=def->zap_code;
	thing->remove_code=def->remove_code;
	thing->nshar=def->sharable;
	thing->version=def->version;
	thing->thing=(Thing)thingbase;

	/* link in the new thing */
	err=THINGLink(thing);
	if (err)
	{
		SYSRelease(thing);
		MEMRelease(thingbase);
	}
	return err;
}

Error THINGCreate(ThingDefinition *def)
{
	return DoCreate(def, Create);
}

Error THINGCreateBlock(ThingDefinition *def)
{
	return DoCreate(def, CreateBlock);
}

static Error DoAppend(ThingDefinition *def, Error (*create)(ThingDefinition *def, Extension **base))
{
	Error err;
	Extension *thingbase;
	Extension *orgthing;
	ThingHeader *thing;

	err=create(def,&thingbase); 	/* create the thing itself */
	if (err) return err;

	err=THINGUse(def->name, 0, (Thing *)&orgthing, NULL);
	if (err) return err;

	/* now find the last extension in the list */
	while (orgthing->next)
	{
		orgthing=(Extension *)(((char *)orgthing)+orgthing->next);
	}

	/* now link the new list in as the next extensions */
	orgthing->flag|=THING_LIST;
	orgthing->next=(char *)thingbase-(char *)orgthing;

	THINGFree(def->name);
	return ERR_OK;
}

Error THINGAppend(ThingDefinition *def)
{
	return DoAppend(def, Create);
}

Error THINGAppendBlock(ThingDefinition *def)
{
	return DoAppend(def, CreateBlock);
}

Error THINGCall(char *name, int extension, int paramcount, ...)
{
	Error err;
	char *thing;

	err=THINGUse(name, extension, (Thing *)&thing, NULL);
	if (err) return err;
	err=ROUTCallList((Error (*)())(thing+THING_CODE), &paramcount);
	THINGFree(name);
	return err;
}

Error THINGCallBlock(char *name, int extension, void *paramblock)
{
	Error err;
	void *thing;

	err=THINGUse(name, 0, &thing, NULL);
	if (err) return err;
	err=THINGCallBlockIn(thing, extension, paramblock);
	THINGFree(name);
	return err;
}

Error THINGCreateAny(ThingDefinition *def, Thing base)
{
	Error err;
	ThingHeader *thing, *thing2;

	/* now allocate and fill in the ThingHeader */
	err=SYSAllocate(sizeof(ThingHeader),def->owner,&thing2);
	if (err) return err;
	thing=thing2;	/* to allow use of register - stupid compiler */
	thing->namelen=min(THING_MAXNAME,strlen(def->name));
	MEMCopy(THING_MAXNAME,def->name,thing->name);
	thing->use_code=def->use_code;
	thing->free_code=def->free_code;
	thing->zap_code=def->zap_code;
	thing->remove_code=def->remove_code;
	thing->nshar=def->sharable;
	thing->version=def->version;
	thing->thing=base;

	/* link in the new thing */
	err=THINGLink(thing);
	if (err)
	{
		SYSRelease(thing);
	}
	return err;
}
