/*
	utility program for PROGS SMS C library

	program to create a configuration file for a C program

	The program has two parameters : the input file with the configuration
	info and the name of the output file (input for as68, _s).
	When channels are passed, the last one is used to display warnings etc.
	When a file with the same name as the output file already exists, it will
	be overwritten.

	The program will generate a config level 2 config block.

	The input file has the following format (lines are limited to 120 chars) :

	- the first line should start with "CONFIG" and then the program name and
	version. (Parameters are handled in the same way as command line
	parameters - see STRBreak).

	- empty lines and lines starting with a semicolon or percentage character
	(";" or "%") are discarded as comments.

	- All lines starting with '"' are treated as the description for the
	following item. This can be multiple lines (each preceded by a quote).

	- "STRING" followed by variable name, max length and default parameters.

	- "FONT" followed by variable name and default.

	- "DRIVER" followed by variable name and default.

	- "DIRECTORY" followed by a variable name and default.

	- "FILENAME" followed by a varaible name and default.

	- "ID" followed by a string. Only the first three letters will be used as the
	top three bytes of the item id. The bottom byte will be initialised to zero
	and incremented by one for each of the following items. When the string
	not contain three characters, then local id's will be generated.

	- "SKIP" - this will generate a skipped item - which can be important for
	backward compatibility with old config blocks, to replace no longer used
	items and not replace the id for that item by something else.

	- "BYTE" followed by variable name, default, minimum and maximum value.
	(These values are treated as unsigned). (the variable will be a char.

	- "SHORT" followed by variable name, default, minimum and maximum value.
	(These values are treated as unsigned).

	- "INT" followed by variable name, default, minimum and maximum value.
	(These values are treated as unsigned).

	- "SELECTION" followed by name. The actual possibilities are displayed on
	the following lines, which should start by one or more spaces, then the
	code value (0..255) and then the description (one word or between double
	quotes. The list is ended by a comment or empty line.

	written by Joachim Van der Auwera for PROGS
	(c) 1995 by PROGS, Professional & Graphical Software
	started 29 October 1995
	01.11.1995 first fully working version
	06.01.1996 added FILENAME and DIRECTORY items
			   a default empty string ("") is now possible
	10.06.1996 now special tag for filename and directory in ProWesS
*/

#include "sms_h"
#include "io_h"
#include "str_h"
#include "job_h"
#include "win_h"
#include "PROforma_h"

#define out(x) IOPutString(outch,x,NULL)
#define report(x) IOPutString(rep,x,NULL)
#define catch(x) if (err=(x)) return err; else
#define MAX_LINE_LENGTH 120

typedef struct _cfg_item {
	struct _cfg_item *next;
	int id, type, item;
	char post;
	} cfg_item;

static WindowDescriptor report_window={ 350,130,80,60,6,6,0,1,2,7 };

Error init(char *param, short chcount, Channel *channels)
{
	Error err;
	Channel in,outch,rep;
	char **params;
	char line[MAX_LINE_LENGTH];
	char prgname[JOB_MAXNAME];
	char prgvers[JOB_MAXNAME];
	Chunks cfg_items;
	cfg_item *items=NULL, *prev_item=NULL;
	int mask_id=0, local_id=0, counter=1;

	if (chcount)
		rep=channels[chcount-1];
	else
		WINOpenFull(&report_window,&rep);

	/* open input and output files */

	err=STRBreak(param,&params);
	if (err) return err;
	if (params[0]==NULL || params[1]==NULL)
	{
		report("wrong parameters, need \"input output\" file");
		return ERR_IPAR;
	}

	catch(MEMUserInit(sizeof(cfg_item),20,&cfg_items));

	err=IOOpenPath(params[0],OPEN_READ,PATH_DATA,&in);
	if (err) { report("can't open input"); return err; }
	err=IOOpenPath(params[1],OPEN_OVER,PATH_DATA,&outch);
	if (err) { report("can't open output"); return err; }
	MEMRelease(params);

	/* initialise the output file - this means writing the post processing */
	/* routine for strings (which adds the '\0') */
	catch(out("; machine generated config file by 'mkconfig'"));
	catch(out("; defines the config block which is part of the file"));
	catch(out("; and includes all the variables which can be configured"));
	catch(out("\n.data"));
	catch(out("Post:"));
	catch(out("    move.w  2(a0),d2"));
	catch(out("    clr.b   4(a0,d2.w)"));
	catch(out("    moveq   #0,d1"));
	catch(out("    moveq   #0,d0"));
	catch(out("    rts\n"));
	catch(out("desc1:"));
	catch(out("    dc.w    endd1-desc1-3"));

	/* start by reading first line which contains the prg name and version */
	/* these have to be stored temporary because writing the actual config */
	/* block is done later */

	err=IOGetString(in,line,MAX_LINE_LENGTH);
	if (err) { report("empty input file ??"); return err; }

	catch(err=STRBreak(line,&params));
	if (params[0]==NULL || !STRSameCI("CONFIG",params[0]))
	{
		report("first line should start with \"CONFIG\"");
		return err;
	}
	if (params[1]==NULL)
	{
		report("CONFIG needs at least a program name");
		return ERR_IPAR;
	}
	STRCopySafe(params[1],prgname,JOB_MAXNAME);
	if (params[2]==NULL)
		prgvers[0]='\0';
	else
		STRCopySafe(params[2],prgvers,JOB_MAXNAME);
	MEMRelease(params);

	/* now process the entire config file */
	/* this will write all strings etc to the output file, but the actual */
	/* config block is only written afterwards */

	while(!(err=IOGetString(in,line,MAX_LINE_LENGTH)))
	{
		cfg_item *item;
		char Line[MAX_LINE_LENGTH];
		int length;

		/* skip all comment lines */
		if (line[0]=='\0' || line[0]==';' || line[0]=='%') continue;

		if (line[0]=='"')
		{
			STRFormatPut(Line,"    dc.b    \"%s\"",&line[1]);
			catch(out(Line));
			catch(out("    dc.b    10"));
			continue;
		}

		catch(STRBreak(line,&params));
		if (params[0]==NULL) continue;
		if (STRSameCI("ID",params[0]))
		{
			if (params[1][0]==0)
			{
				report("no id in SKIP command");
				return ERR_IPAR;
			}
			mask_id=params[1][0];
			if (params[1][1]!=0)
			{
				mask_id=(mask_id<<8)+params[1][1];
				if (params[1][2]!=0)
					mask_id=(mask_id<<8)+params[1][2];
			}
			mask_id<<=8;
			local_id=1;
			MEMRelease(params);
			continue;
		}
		local_id++;
		if (local_id>255)
		{
			report("can't have more than 255 items for one ID");
			return ERR_IPAR;
		}
		if (STRSameCI("SKIP",params[0]))
		{
			MEMRelease(params);
			continue;
		}

		STRFormatPut(Line,"endd%d:\n.even\nattr%d:",counter,counter);
		catch(out(Line));
		catch(MEMUserAllocate(cfg_items,&item));
		item->id=mask_id+local_id;
		item->next=NULL;
		if (prev_item==NULL)
			items=item;
		else
			prev_item->next=item;
		prev_item=item;
		item->item=counter;
		item->post=FALSE;

		if (STRSameCI("STRING",params[0]))
		{
			catch(out("    dc.w    0"));
			STRFormatGet(params[2],NULL,"%d",&length);
strput: 	item->type=0;
			item->post=TRUE;
			STRFormatPut(Line,".globl _%s",params[1]);
			catch(out(Line));
			STRFormatPut(Line,"item%d:\n    dc.w    %d",counter,length-1);
			catch(out(Line));
			STRFormatPut(Line,"    dc.w    %d",strlen(params[3]));
			catch(out(Line));
			if (strlen(params[3]))
				STRFormatPut(Line,"_%s:\n    dc.b    '%s'",params[1],params[3]);
			else
				STRFormatPut(Line,"_%s:",params[1]);
			catch(out(Line));
			STRFormatPut(Line,"    ds.b    %d",length-strlen(params[3]));
			catch(out(Line));
		}
		else if (STRSameCI("FONT",params[0]))
		{
			catch(out("    dc.w    $2100"));
			length=PF_MAXFONTNAME;
			params[3]=params[2];
			goto strput;
		}
		else if (STRSameCI("DRIVER",params[0]))
		{
			catch(out("    dc.w    $2200"));
			length=PF_MAXDRIVERNAME;
			params[3]=params[2];
			goto strput;
		}
		else if (STRSameCI("DIRECTORY",params[0]))
		{
			catch(out("    dc.w    $2300"));
			length=IO_MAXDIRECTORY;
			params[3]=params[2];
			goto strput;
		}
		else if (STRSameCI("FILENAME",params[0]))
		{
			catch(out("    dc.w    $2400"));
			length=IO_MAXFULLNAME;
			params[3]=params[2];
			goto strput;
		}
		else if (STRSameCI("BYTE",params[0]))
		{
			STRFormatPut(Line,"    dc.b    %s,%s",params[3],params[4]);
			catch(out(Line));
			item->type=8;
			STRFormatPut(Line,".globl _%s\n_%s:",params[1],params[1]);
			catch(out(Line));
			STRFormatPut(Line,"item%d:\n    dc.b    %s",counter,params[2]);
			catch(out(Line));
		}
		else if (STRSameCI("SHORT",params[0]))
		{
			STRFormatPut(Line,"    dc.w    %s,%s",params[3],params[4]);
			catch(out(Line));
			item->type=10;
			STRFormatPut(Line,".globl _%s\n_%s:",params[1],params[1]);
			catch(out(Line));
			STRFormatPut(Line,"item%d:\n    dc.w    %s",counter,params[2]);
			catch(out(Line));
		}
		else if (STRSameCI("INT",params[0]))
		{
			STRFormatPut(Line,"    dc.l    %s,%s",params[3],params[4]);
			catch(out(Line));
			item->type=12;
			STRFormatPut(Line,".globl _%s\n_%s:",params[1],params[1]);
			catch(out(Line));
			STRFormatPut(Line,"item%d:\n    dc.l    %s",counter,params[2]);
			catch(out(Line));
		}
		else if (STRSameCI("SELECTION",params[0]))
		{
			char **choices;
			char process_choices=TRUE;

			/* start by procssing the next lines with the choices */
			while(process_choices)
			{
				err=IOGetString(in,line,MAX_LINE_LENGTH);
				if (err && err!=ERR_EOF)
					return err;
				if (err || line[0]=='\0' || line[0]==';' || line[0]=='%')
				{
					process_choices=FALSE;
					continue;
				}
				catch(STRBreak(line,&choices));
				STRFormatPut(Line,"    dc.b    %s,0",choices[0]);
				catch(out(Line));
				STRFormatPut(Line,"    dc.w    %d",strlen(choices[1]));
				catch(out(Line));
				STRFormatPut(Line,"    dc.b    '%s'\n.even",choices[1]);
				catch(out(Line));
				MEMRelease(choices);
			}
			catch(out("    dc.w    -1"));
			item->type=4;
			STRFormatPut(Line,".globl _%s\n_%s:",params[1],params[1]);
			catch(out(Line));
			STRFormatPut(Line,"item%d:\n    dc.b    %s",counter,params[2]);
			catch(out(Line));
		}
		MEMRelease(params);
		counter++;
		STRFormatPut(Line,".even\ndesc%d:",counter);
		catch(out(Line));
		STRFormatPut(Line,"    dc.w    endd%d-desc%d-3",counter,counter);
		catch(out(Line));
	}
	if (err!=ERR_EOF) { report("problem while reading file"); return err; }
	IOClose(in);

	/* write the final part of the config file - the actual config block */

	/* to write th header we have to make sure that this executable does not */
	/* become configurable */
	STRFormatPut(line,".even\n    dc.b    '%s%s'","<<QC","FX>>02");
	catch(out(line));
	STRFormatPut(line,"    dc.w    %d",strlen(prgname));
	catch(out(line));
	STRFormatPut(line,"    dc.b    '%s'\n.even",prgname);
	catch(out(line));
	STRFormatPut(line,"    dc.w    %d",strlen(prgvers));
	catch(out(line));
	STRFormatPut(line,"    dc.b    '%s'\n.even",prgvers);
	catch(out(line));

	for ( ; items ; items=items->next )
	{
		STRFormatPut(line,"    dc.l    %d",items->id);
		catch(out(line));
		STRFormatPut(line,"    dc.b    %d,0",items->type);
		catch(out(line));
		STRFormatPut(line,"h%d:\n    dc.w    item%d-h%d",items->item,items->item,items->item);
		catch(out(line));
		if (items->post)
		{
			STRFormatPut(line,"    dc.w    0,Post-h%d-4",items->item);
			catch(out(line));
		}
		else
			catch(out("    dc.w    0,0"));
		STRFormatPut(line,"    dc.w    desc%d-h%d-6",items->item,items->item);
		catch(out(line));
		STRFormatPut(line,"    dc.w    attr%d-h%d-8",items->item,items->item);
		catch(out(line));
	}

	catch(out("    dc.w    -1\n.end"));
	return ERR_OK;
}
